#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#ifdef MEMWATCH
#include "memwatch.h"
#endif

#define _MALLOC_C_
#include "malloc.h"

// Ǝ}l[WgpꍇÃRgOĂB
#define USE_MEMMGR

// fobO[hiG[`FbN̋pAʏ^cł͂߂łȂj
// #define DEBUG_MEMMGR

#ifndef USE_MEMMGR

void* aMalloc_( size_t size, const char *file, int line, const char *func )
{
	void *ret;
	
	// printf("%s:%d: in func %s: malloc %d\n",file,line,func,size);
#ifdef MEMWATCH
	ret=mwMalloc(size,file,line);
#else
	ret=malloc(size);
#endif
	if(ret==NULL){
		printf("%s:%d: in func %s: malloc error out of memory!\n",file,line,func);
		exit(1);

	}
	return ret;
}

void* aCalloc_( size_t num, size_t size, const char *file, int line, const char *func )
{
	void *ret;
	
	// printf("%s:%d: in func %s: calloc %d %d\n",file,line,func,num,size);
#ifdef MEMWATCH
	ret=mwCalloc(num,size,file,line);
#else
	ret=calloc(num,size);
#endif
	if(ret==NULL){
		printf("%s:%d: in func %s: calloc error out of memory!\n",file,line,func);
		exit(1);

	}
	return ret;
}

void* aRealloc_( void *p, size_t size, const char *file, int line, const char *func )
{
	void *ret;
	
	// printf("%s:%d: in func %s: realloc %p %d\n",file,line,func,p,size);
#ifdef MEMWATCH
	ret=mwRealloc(p,size,file,line);
#else
	ret=realloc(p,size);
#endif
	if(ret==NULL){
		printf("%s:%d: in func %s: realloc error out of memory!\n",file,line,func);
		exit(1);

	}
	return ret;
}

void* aStrdup_( const void *p, const char *file, int line, const char *func )
{
	void *ret;
	
	// printf("%s:%d: in func %s: strdup %p\n",file,line,func,p);
#ifdef MEMWATCH
	ret=mwStrdup(p,file,line);
#else
	ret=strdup(p);
#endif
	if(ret==NULL){
		printf("%s:%d: in func %s: strdup error out of memory!\n",file,line,func);
		exit(1);

	}
	return ret;
}

void aFree_( void *p, const char *file, int line, const char *func )
{
	// printf("%s:%d: in func %s: free %p\n",file,line,func,p);
#ifdef MEMWATCH
	mwFree(p,file,line);
#else
	free(p);
#endif
}

int do_init_memmgr(const char* file) {
	return 0;
}

double memmgr_usage(void) {
	return 0;
}

#else /* USE_MEMMGR */

/*
 * }l[W
 *     malloc , free ̏Iɏo悤ɂ́B
 *     GȏsĂ̂ŁA኱dȂ邩܂B
 *
 * f[^\Ȃǁił܂^^; j
 *     E𕡐́uubNvɕāAɃubN𕡐́ujbgv
 *       ɕĂ܂Bjbg̃TCÝAPubN̗eʂ𕡐ɋϓz
 *       ̂łBƂ΁APjbg32KB̏ꍇAubNP32Bytẽ
 *       jbgA1024W܂ďoĂA64Bytẽjbg 512W܂
 *       oĂ肵܂Bipadding,unit_head j
 *
 *     EubNm̓NXg(block_prev,block_next) łȂATC
 *       YubNmNXg(hash_prev,hash_nect) ł
 *       Ă܂BɂAsvƂȂ̍ėpIɍs܂B
 */

/* ubÑACg */
#define BLOCK_ALIGNMENT1	16
#define BLOCK_ALIGNMENT2	64

/* ubNɓf[^ */
#define BLOCK_DATA_COUNT1	128
#define BLOCK_DATA_COUNT2	608

/* ubN̑傫: 16*128 + 64*576 = 40KB */
#define BLOCK_DATA_SIZE1	( BLOCK_ALIGNMENT1 * BLOCK_DATA_COUNT1 )
#define BLOCK_DATA_SIZE2	( BLOCK_ALIGNMENT2 * BLOCK_DATA_COUNT2 )
#define BLOCK_DATA_SIZE		( BLOCK_DATA_SIZE1 + BLOCK_DATA_SIZE2 )

/* xɊmۂubN̐B */
#define BLOCK_ALLOC		104

/* ubN */
struct block {
	struct block* block_next;		/* Ɋmۂ̈ */
	struct block* unfill_prev;		/* ̖܂ĂȂ̈ */
	struct block* unfill_next;		/* ̖܂ĂȂ̈ */
	unsigned short unit_size;		/* jbg̑傫 */
	unsigned short unit_hash;		/* jbg̃nbV */
	unsigned short unit_count;		/* jbǧ */
	unsigned short unit_used;		/* gpjbg */
	unsigned short unit_unfill;		/* gpjbg̏ꏊ */
	unsigned short unit_maxused;	/* gpjbg̍ől */
	char   data[ BLOCK_DATA_SIZE ];
};

struct unit_head {
	struct block   *block;
	const  char*   file;
	unsigned short line;
	unsigned short size;
	long           checksum;
};

static struct block* hash_unfill[BLOCK_DATA_COUNT1 + BLOCK_DATA_COUNT2 + 1];
static struct block* block_first, *block_last, block_head;

/* g񂹂Ȃ̈p̃f[^ */
struct unit_head_large {
	size_t                  size;
	struct unit_head_large* prev;
	struct unit_head_large* next;
	struct unit_head        unit_head;
};

static struct unit_head_large *unit_head_large_first = NULL;

static struct block* block_malloc(unsigned short hash);
static void          block_free(struct block* p);
static void          memmgr_warning(const char* format,...);
static unsigned int  memmgr_usage_bytes;

#define block2unit(p, n) ((struct unit_head*)(&(p)->data[ p->unit_size * (n) ]))

#ifdef DEBUG_MEMMGR
#define memmgr_assert(v) do { if(!(v)) { memmgr_warning("memmgr_assert: " #v "\n"); } } while(0)
#else
#define memmgr_assert(v)
#endif

static unsigned short size2hash( size_t size ) {
	if( size <= BLOCK_DATA_SIZE1 ) {
		return (size + BLOCK_ALIGNMENT1 - 1) / BLOCK_ALIGNMENT1;
	} else {
		return (size - BLOCK_DATA_SIZE1 + BLOCK_ALIGNMENT2 - 1) / BLOCK_ALIGNMENT2
				+ BLOCK_DATA_COUNT1;
	}
}

static size_t hash2size( unsigned short hash ) {
	if( hash <= BLOCK_DATA_COUNT1) {
		return hash * BLOCK_ALIGNMENT1;
	} else {
		return (hash - BLOCK_DATA_COUNT1) * BLOCK_ALIGNMENT2 + BLOCK_DATA_SIZE1;
	}
}

void* aMalloc_(size_t size, const char *file, int line, const char *func) {
	struct block *block;
	short size_hash = size2hash( size );
	struct unit_head *head;
	
	if(size == 0) {
		return NULL;
	}
	memmgr_usage_bytes += size;

	/* ubN𒴂̈̊mۂɂ́Amalloc() p */
	/* ̍ہAunit_head.block  NULL ċʂ */
	if(hash2size(size_hash) > BLOCK_DATA_SIZE - sizeof(struct unit_head)) {
#ifdef MEMWATCH
		struct unit_head_large* p = (struct unit_head_large*)mwMalloc(sizeof(struct unit_head_large) + size,file,line);
#else
		struct unit_head_large* p = (struct unit_head_large*)malloc(sizeof(struct unit_head_large) + size);
#endif
		if(p != NULL) {
			p->size            = size;
			p->unit_head.block = NULL;
			p->unit_head.size  = 0;
			p->unit_head.file  = file;
			p->unit_head.line  = line;
			if(unit_head_large_first == NULL) {
				unit_head_large_first = p;
				p->next = NULL;
				p->prev = NULL;
			} else {
				unit_head_large_first->prev = p;
				p->prev = NULL;
				p->next = unit_head_large_first;
				unit_head_large_first = p;
			}
			*(long*)((char*)p + sizeof(struct unit_head_large) - sizeof(long) + size) = 0xdeadbeaf;
			return (char *)p + sizeof(struct unit_head_large) - sizeof(long);
		} else {
			printf("MEMMGR::memmgr_alloc failed.\n");
			exit(1);
		}
	}

	/* TCỸubNmۂĂȂAVɊmۂ */
	if(hash_unfill[size_hash]) {
		block = hash_unfill[size_hash];
	} else {
		block = block_malloc(size_hash);
	}

	if( block->unit_unfill == 0xFFFF ) {
		// freeςݗ̈悪cĂȂ
		memmgr_assert(block->unit_used <  block->unit_count);
		memmgr_assert(block->unit_used == block->unit_maxused);
		head = block2unit(block, block->unit_maxused);
		block->unit_used++;
		block->unit_maxused++;
	} else {
		head = block2unit(block, block->unit_unfill);
		block->unit_unfill = head->size;
		block->unit_used++;
	}

	if( block->unit_unfill == 0xFFFF && block->unit_maxused >= block->unit_count) {
		// jbggʂ̂ŁAunfillXg폜
		if( block->unfill_prev == &block_head) {
			hash_unfill[ size_hash ] = block->unfill_next;
		} else {
			block->unfill_prev->unfill_next = block->unfill_next;
		}
		if( block->unfill_next ) {
			block->unfill_next->unfill_prev = block->unfill_prev;
		}
		block->unfill_prev = NULL;
	}

	head->block = block;
	head->file  = file;
	head->line  = line;
	head->size  = size;
	*(long*)((char*)head + sizeof(struct unit_head) - sizeof(long) + size) = 0xdeadbeaf;
	return (char *)head + sizeof(struct unit_head) - sizeof(long);
};

void* aCalloc_(size_t num, size_t size, const char *file, int line, const char *func ) {
	void *p = aMalloc_(num * size,file,line,func);
	memset(p,0,num * size);
	return p;
}

void* aRealloc_(void *memblock, size_t size, const char *file, int line, const char *func ) {
	size_t old_size;
	if(memblock == NULL) {
		return aMalloc_(size,file,line,func);
	}

	old_size = ((struct unit_head *)((char *)memblock - sizeof(struct unit_head) + sizeof(long)))->size;
	if( old_size == 0 ) {
		old_size = ((struct unit_head_large *)((char *)memblock - sizeof(struct unit_head_large) + sizeof(long)))->size;
	}
	if(old_size > size) {
		// TCYk -> ̂܂ܕԂi蔲j
		return memblock;
	}  else {
		// TCYg
		void *p = aMalloc_(size,file,line,func);
		if(p != NULL) {
			memcpy(p,memblock,old_size);
		}
		aFree_(memblock,file,line,func);
		return p;
	}
}

void* aStrdup_(const void* string, const char *file, int line, const char *func ) {
	if(string == NULL) {
		return NULL;
	} else {
		size_t  len = strlen(string);
		char    *p  = (char *)aMalloc_(len + 1,file,line,func);
		memcpy(p,string,len+1);
		return p;
	}
}

void aFree_(void *ptr, const char *file, int line, const char *func ) {
	struct unit_head *head;

	if(ptr == NULL) return; 
	head = (struct unit_head *)((char *)ptr - sizeof(struct unit_head) + sizeof(long));
	if(head->size == 0) {
		/* malloc() ŒɊmۂꂽ̈ */
		struct unit_head_large *head_large = (struct unit_head_large *)((char *)ptr - sizeof(struct unit_head_large) + sizeof(long));
		if(
			*(long*)((char*)head_large + sizeof(struct unit_head_large) - sizeof(long) + head_large->size)
			!= 0xdeadbeaf)
		{
			memmgr_warning("memmgr: args of aFree is overflowed pointer %s line %d\n",file,line);
		} else {
			head->size = -1;
			if(head_large->prev) {
				head_large->prev->next = head_large->next;
			} else {
				unit_head_large_first  = head_large->next;
			}
			if(head_large->next) {
				head_large->next->prev = head_large->prev;
			}
			memmgr_usage_bytes -= head_large->size;
			free(head_large);
		}
	} else {
		/* jbg */
		struct block *block = head->block;
		if( (char*)head - (char*)block > sizeof(struct block) ) {
			memmgr_warning("memmgr: args of aFree is invalid pointer %s line %d\n",file,line);
		} else if(head->block == NULL) {
			memmgr_warning("memmgr: args of aFree is freed pointer %s line %d\n",file,line);
		} else if(*(long*)((char*)head + sizeof(struct unit_head) - sizeof(long) + head->size) != 0xdeadbeaf) {
			memmgr_warning("memmgr: args of aFree is overflowed pointer %s line %d\n",file,line);
		} else {
			memmgr_usage_bytes -= head->size;
			head->block         = NULL;
#ifdef DEBUG_MEMMGR
			memset(ptr, 0xdd, head->size);
#endif
			memmgr_assert( block->unit_used > 0 );
			if(--block->unit_used == 0) {
				/* ubN̉ */
				block_free(block);
			} else {
				if( block->unfill_prev == NULL) {
					// unfill Xgɒǉ
					if( hash_unfill[ block->unit_hash ] ) {
						hash_unfill[ block->unit_hash ]->unfill_prev = block;
					}
					block->unfill_prev = &block_head;
					block->unfill_next = hash_unfill[ block->unit_hash ];
					hash_unfill[ block->unit_hash ] = block;
				}
				head->file     = NULL;
				head->size     = block->unit_unfill;
				block->unit_unfill = (unsigned short)(((unsigned long)head - (unsigned long)block->data) / block->unit_size);
			}
		}
	}
}

/* ubNmۂ */
static struct block* block_malloc(unsigned short hash) {
	int    i;
	struct block *p;
	if(hash_unfill[0] != NULL) {
		/* ubNp̗̈͊mۍς */
		p = hash_unfill[0];
		hash_unfill[0] = hash_unfill[0]->unfill_next;
	} else {
		/* ubNp̗̈VɊmۂ */
		p = (struct block*)malloc(sizeof(struct block) * (BLOCK_ALLOC) );
		if(p == NULL) {
			printf("MEMMGR::block_alloc failed.\n");
			exit(1);
		}
		if(block_first == NULL) {
			/* m */
			block_first   = p;
		} else {
			block_last->block_next = p;
		}
		block_last = &p[BLOCK_ALLOC - 1];
		block_last->block_next = NULL;
		/* ubNA */
		for(i=0;i<BLOCK_ALLOC;i++) {
			if(i != 0) {
				// p[0] ͂ꂩĝŃNɂ͉Ȃ
				p[i].unfill_next = hash_unfill[0];
				hash_unfill[0]   = &p[i];
				p[i].unfill_prev = NULL;
			}
			if(i != BLOCK_ALLOC -1) {
				p[i].block_next = &p[i+1];
			}
		}
	}

	// unfill ɒǉ
	memmgr_assert(hash_unfill[ hash ] == NULL);
	hash_unfill[ hash ] = p;
	p->unfill_prev  = &block_head;
	p->unfill_next  = NULL;
	p->unit_size    = hash2size( hash ) + sizeof(struct unit_head);
	p->unit_hash    = hash;
	p->unit_count   = BLOCK_DATA_SIZE / p->unit_size;
	p->unit_used    = 0;
	p->unit_unfill  = 0xFFFF;
	p->unit_maxused = 0;
	return p;
}

static void block_free(struct block* p) {
	if( p->unfill_prev ) {
		if( p->unfill_prev == &block_head) {
			hash_unfill[ p->unit_hash ] = p->unfill_next;
		} else {
			p->unfill_prev->unfill_next = p->unfill_next;
		}
		if( p->unfill_next ) {
			p->unfill_next->unfill_prev = p->unfill_prev;
		}
		p->unfill_prev = NULL;
	}

	p->unfill_next = hash_unfill[0];
	hash_unfill[0] = p;
}

static char memmer_logfile[128];

static void memmgr_warning(const char* format,...) {
	FILE *fp = fopen(memmer_logfile,"a");
	va_list ap;
	va_start(ap,format);

	if(fp) {
		vfprintf(fp,format,ap);
		fclose(fp);
	}
	vprintf(format,ap);
	va_end(ap);
}

static FILE* memmgr_log(void) {
	FILE *fp = fopen(memmer_logfile,"a");
	if(!fp) { fp = stdout; }
	fprintf(fp,"memmgr: memory leaks found\n");
	return fp;
}

static void memmer_exit(void) {
	FILE *fp = NULL;
//	int i;
	int count = 0;
	struct block *block = block_first;
	struct unit_head_large *large = unit_head_large_first;
	while(block) {
		if(block->unfill_prev) {
			int i;
			if(!fp) { fp = memmgr_log(); }
			for(i=0;i< block->unit_maxused;i++) {
				struct unit_head *head = block2unit(block, i);
				if(head->block != NULL) {
					fprintf(
						fp,"%04d : %s line %d size %d\n",++count,
						head->file,head->line,head->size
					);
				}
			}
		}
		block = block->block_next;
	}
	while(large) {
		if(!fp) { fp = memmgr_log(); }
		fprintf(
			fp,"%04d : %s line %d size %d\n",++count,
			large->unit_head.file,
			large->unit_head.line,large->unit_head.size
		);
		large = large->next;
	}
	if(!fp) {
		printf("memmgr: no memory leaks found.\n");
	} else {
		printf("memmgr: memory leaks found.\n");
	}
}

int do_init_memmgr(const char* file) {
	sprintf(memmer_logfile,"%s.log",file);
	atexit(memmer_exit);
	printf("memmgr: initialised: %s\n",memmer_logfile);
	return 0;
}

double memmgr_usage(void) {
	return memmgr_usage_bytes / 1024.0;
}

#endif
